import re
from collections.abc import Generator

import pytest
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver

from myst_nb_json import CLS_COLLAPSED, JsonMimeRenderPlugin


@pytest.fixture
def plugin() -> JsonMimeRenderPlugin:
    return JsonMimeRenderPlugin()


JSON_DATA = {
    "key1": "value1",
    "key2": {
        "key21": "str",
        "key22": 42,
        "key23": 3.14,
        "key24": True,
        "key25": False,
        "key26": None,
        "key27": {},
        "key28": [],
    },
    "key3": ["str", 42, 3.14, True, False, None, {}, []],
}


@pytest.fixture
def browser() -> Generator[WebDriver]:
    # Set up headless Chrome options
    chrome_options = Options()
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--disable-gpu")

    # Initialize the WebDriver
    driver: WebDriver = webdriver.Chrome(options=chrome_options)
    yield driver
    driver.quit()


def test_interactive_collapse_expand(plugin: JsonMimeRenderPlugin, browser: WebDriver):
    # Load the HTML string
    html_content = plugin.html(JSON_DATA, root="test", expanded=True)
    # Hash (in CSS color) causes problems for Selenium Chrome webdriver
    html_content = re.sub("#", "&num;", html_content)
    browser.get("data:text/html;charset=utf-8," + html_content)

    # Wait for the JavaScript to execute
    browser.implicitly_wait(2)

    text_to_click = "key2"
    element_to_click = browser.find_element(By.XPATH, f"//*[contains(text(), '{text_to_click}')]")
    value_element = element_to_click.find_element(By.XPATH, "following-sibling::*[1]")
    li_element = element_to_click.find_element(By.XPATH, "ancestor::li[1]")

    assert CLS_COLLAPSED not in li_element.get_attribute("class")
    assert value_element.value_of_css_property("display") != "none"

    # Collapse
    element_to_click.click()
    assert CLS_COLLAPSED in li_element.get_attribute("class")
    assert value_element.value_of_css_property("display") == "none"

    # Expand
    element_to_click.click()
    assert CLS_COLLAPSED not in li_element.get_attribute("class")
    assert value_element.value_of_css_property("display") != "none"



import pytest

from tests.conftest import clean_doctree, file_regression, get_test_path, sphinx_run  # noqa: F401


@pytest.mark.sphinx_params("json_output.ipynb", conf={"nb_execution_mode": "force"})
def test_render_json_output(
    sphinx_run,  # noqa: F811
    clean_doctree,  # noqa: F811
    file_regression,  # noqa: F811
    get_test_path,  # noqa: F811
):
    """Test that a cell with JSON output is rendered with the MIME type plugin"""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("json_output"))
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")



import re

import pytest

from myst_nb_json import JsonMimeRenderPlugin


@pytest.fixture
def plugin() -> JsonMimeRenderPlugin:
    return JsonMimeRenderPlugin()


def _strip_xml_attributes(string: str) -> str:
    return re.sub(pattern=r"<(?P<tag>\w+)(?:\s+[^>]*)>", repl=r"<\g<tag>>", string=string)


def _strip_xml_whitespace(string: str) -> str:
    return re.sub(pattern=r">\s+<", repl="><", string=string)


def _strip_xml_tags(string: str) -> str:
    return re.sub(pattern=r"<\/?\w+[^>]*>", repl="", string=string)


CHAR_to_HTML_ENTITY = {
    ">": "&gt;",
    "<": "&lt;",
    "&": "&amp;",
    # These also cause problems, although they should not strictly need escaping
    "[": "&lbrack;",
    "]": "&rbrack;",
    "{": "&lbrace;",
    "}": "&rbrace;",
    '"': "&quot;",
    # Hash (in CSS color) causes problems for Selenium Chrome webdriver
    "#": "&num;",
}
HTML_ENTITY_TO_CHAR = {v: k for k, v in CHAR_to_HTML_ENTITY.items()}


def _html_unescape(string: str) -> str:
    return re.sub(
        pattern=r"&\w+;",
        repl=lambda m: HTML_ENTITY_TO_CHAR.get(m.group(), m.group()),
        string=string,
    )


@pytest.mark.parametrize(
    ("value", "expected_xml"),
    [
        ("abc", '<span>"abc"</span>'),
        (42, "<span>42</span>"),
        (3.14, "<span>3.14</span>"),
        (True, "<span>true</span>"),
        (False, "<span>false</span>"),
        (None, "<span>null</span>"),
    ],
)
def test_scalar_value(plugin: JsonMimeRenderPlugin, value, expected_xml: str):
    actual = "".join(plugin.scalar_value(value))
    # It should produce the expected HTML structure
    assert _strip_xml_whitespace(_strip_xml_attributes(actual)) == expected_xml


@pytest.mark.parametrize(
    ("value", "expected_xml", "expected_text"),
    [
        ([], "<div><span>&lbrack;</span><ul></ul><span>&rbrack;</span></div>", "[]"),
        (
            ["a", "b"],
            '<div><span>&lbrack;</span><ul><li><span>"a"</span><span>, </span></li><li><span>"b"</span></li></ul><span>&rbrack;</span></div>',
            '["a", "b"]',
        ),
    ],
)
def test_list_value(plugin: JsonMimeRenderPlugin, value, expected_xml: str, expected_text: str):
    actual = "".join(plugin.list_value(value))
    # It should produce the expected HTML structure
    assert _strip_xml_whitespace(_strip_xml_attributes(actual)) == expected_xml
    # It should preserve JSON syntax as user-electable text (even if brackets are not displayed)
    assert _html_unescape(_strip_xml_tags(actual)) == expected_text


@pytest.mark.parametrize(
    ("value", "expected_xml", "expected_text"),
    [
        ({}, "<div><span>&lbrace;</span><ul></ul><span>&rbrace;</span></div>", "{}"),
        (
            {"k1": "v1", "k2": "v2"},
            '<div><span>&lbrace;</span><ul><li><span><span>"</span>k1<span>"</span><span>: </span></span><span>"v1"</span><span>, </span></li><li><span><span>"</span>k2<span>"</span><span>: </span></span><span>"v2"</span></li></ul><span>&rbrace;</span></div>',
            '{"k1": "v1", "k2": "v2"}',
        ),
    ],
)
def test_dict_value(plugin: JsonMimeRenderPlugin, value, expected_xml: str, expected_text: str):
    actual = "".join(plugin.dict_value(value))
    # It should produce the expected HTML structure
    assert _strip_xml_whitespace(_strip_xml_attributes(actual)) == expected_xml
    # It should preserve JSON syntax as user-electable text (even if braces and quotes are not displayed)
    assert _html_unescape(_strip_xml_tags(actual)) == expected_text



"""
MimeRenderPlugin for MyST-NB for rendering IPython JSON display type to HTML
"""

__all__ = ["JsonMimeRenderPlugin"]
__version__ = "0.1.3"

import json
from collections.abc import Generator, Mapping, Sequence
from functools import cached_property
from importlib.resources import files
from typing import Union, cast

from docutils import nodes
from myst_nb.core.render import MimeData, MimeRenderPlugin, NbElementRenderer

JAVASCRIPT_FILE_NAME = "myst-nb-json.js"
CSS_FILE_NAME = "myst-nb-json.css"
CLS_COLLAPSIBLE = "myst-nb-json-collapsible"
CLS_COLLAPSED = "myst-nb-json-collapsed"
CLS_HIDDEN = "myst-nb-json-hidden"
CLS_KEY = "myst-nb-json-key"
CLS_UNSELECTABLE = "myst-nb-json-unselectable"
CLS_VALUE = "myst-nb-json-value"

ScalarJsonType = Union[str, int, float, bool, None]

JsonType = Union[Mapping[str, "JsonType"], Sequence["JsonType"], ScalarJsonType]


class JsonMimeRenderPlugin(MimeRenderPlugin):
    mime_priority_overrides = [("*", "application/json", 1)]

    @staticmethod
    def handle_mime(
        renderer: NbElementRenderer, data: MimeData, inline: bool
    ) -> Union[None, list[nodes.Element]]:
        """
        A function that renders a mime type to docutils nodes, or returns None to reject

        Args:
            renderer: A class for rendering notebook elements
            data: The Notebook output data
            inline: Whether the MIME type is displayed inline, e.g. in an interactive Jupyter
                notebook. In the context of MyST-NB, content is rendered to a file, so the renderer
                should not accept inline content.

        Returns:
            A list of docutils nodes, or None if this renderer cannot handle the data
        """
        try:
            if not inline and data.mime_type == "application/json":
                root = data.output_metadata.get(data.mime_type, {}).get("root", "root")
                expanded = data.output_metadata.get(data.mime_type, {}).get("expanded", True)
                plugin = JsonMimeRenderPlugin()
                html = plugin.html(data.content, root=root, expanded=expanded)
                return [nodes.raw(text=html, format="html", classes=["output", "text_html"])]
        except Exception as e:
            import traceback

            print(e)
            traceback.print_exc()
        return None

    def html(self, jsonable: JsonType, root: str = "root", expanded: bool = False) -> str:
        """
        Generate an HTML string for a JSON object

        Args:
            jsonable: A JSON-like Python object
            root: Optional name to display as key for the JSON value
            expanded: Whether to initialize the JSON tree collapsed or expanded

        Returns:
            The component's HTML as string
        """
        return "".join(self.component(jsonable, root=root, expanded=expanded))

    def component(
        self, jsonable: JsonType, root: str = "root", expanded: bool = False
    ) -> Generator[str, None, None]:
        """
        Yield HTML fragment strings for a JSON object

        Args:
            jsonable: A JSON-like Python object
            root: Optional name to display as key for the JSON value
            expanded: Whether to initialize the JSON tree collapsed or expanded

        Yields:
            HTML strings
        """
        yield "<div>"
        yield f"""<div class="{CLS_VALUE}"><ul><li>"""
        yield from self.key(root, collapsible=is_nested(jsonable), selectable=False)
        yield from self.value(jsonable, expanded=expanded)
        yield "</li></ul></div>"
        yield self.style
        yield self.script
        yield "</div>"

    @cached_property
    def script(self) -> str:
        script_str = (files(__package__) / "resources" / JAVASCRIPT_FILE_NAME).read_text()
        return f"<script defer>{script_str}</script>"

    @cached_property
    def style(self) -> str:
        css_str = (files(__package__) / "resources" / CSS_FILE_NAME).read_text()
        return f"""<style>{css_str}</style>"""

    def value(
        self, value: JsonType, expanded: bool = False, with_comma: bool = False
    ) -> Generator[str, None, None]:
        if isinstance(value, Sequence) and not isinstance(value, str) and len(value) > 0:
            yield from self.list_value(value, expanded=expanded, with_comma=with_comma)
        elif isinstance(value, Mapping) and len(value) > 0:
            yield from self.dict_value(value, expanded=expanded, with_comma=with_comma)
        else:
            yield from self.scalar_value(cast(ScalarJsonType, value), with_comma=with_comma)

    def key(
        self, key: Union[str, int], collapsible: bool = False, selectable: bool = True
    ) -> Generator[str, None, None]:
        classes = CLS_KEY
        if collapsible:
            classes += " " + CLS_COLLAPSIBLE
        if not selectable:
            classes += " " + CLS_UNSELECTABLE
        yield f"""<span class="{classes}" tabindex=0>{self.quote}{key}{self.quote}{self.colon}</span>"""

    def list_value(
        self, value: Sequence, expanded: bool = False, with_comma: bool = False
    ) -> Generator[str, None, None]:
        yield f"""<div class="{CLS_VALUE}">{self.bracket_open}<ul>"""
        last_index = len(value) - 1
        for index, val in enumerate(value):
            is_last = index == last_index
            yield f"""<li class="{CLS_COLLAPSED if is_nested(val) and not expanded else ''}">"""
            # yield from self.key(index, collapsible=is_nested(val), selectable=False)
            yield from self.value(val, with_comma=not is_last)
            # It would be nicer to add the comma here, but if value yields a block element,
            # the comma would be an orphan in the next line below the value. To have it in the same
            # line, it needs to be inside the value's block.
            yield "</li>"
        yield f"</ul>{self.bracket_close}{self.comma if with_comma else ''}</div>"

    def dict_value(
        self, value: Mapping, expanded: bool = False, with_comma: bool = False
    ) -> Generator[str, None, None]:
        yield f"""<div class="{CLS_VALUE}">{self.curly_open}<ul>"""
        last_index = len(value.items()) - 1
        for index, (key, val) in enumerate(value.items()):
            is_last = index == last_index
            yield f"""<li class="{CLS_COLLAPSED if is_nested(val) and not expanded else ''}">"""
            yield from self.key(key, collapsible=is_nested(val))
            yield from self.value(val, with_comma=not is_last)
            # It would be nicer to add the comma here.
            yield "</li>"
        yield f"</ul>{self.curly_close}{self.comma if with_comma else ''}</div>"

    def scalar_value(
        self, value: ScalarJsonType, with_comma: bool = False
    ) -> Generator[str, None, None]:
        value_str = json.dumps(value)
        yield f"""<span class="{CLS_VALUE}">{value_str}</span>{self.comma if with_comma else ''}"""

    @cached_property
    def quote(self):
        return f"""<span class="{CLS_HIDDEN}">"</span>"""

    @cached_property
    def colon(self):
        return f"""<span class="{CLS_HIDDEN}">: </span>"""

    @cached_property
    def comma(self):
        return f"""<span class="{CLS_HIDDEN}">, </span>"""

    @cached_property
    def curly_open(self):
        return f"""<span class="{CLS_HIDDEN}">&lbrace;</span>"""

    @cached_property
    def curly_close(self):
        return f"""<span class="{CLS_HIDDEN}">&rbrace;</span>"""

    @cached_property
    def bracket_open(self):
        return f"""<span class="{CLS_HIDDEN}">&lbrack;</span>"""

    @cached_property
    def bracket_close(self):
        return f"""<span class="{CLS_HIDDEN}">&rbrack;</span>"""


def is_nested(jsonable: JsonType) -> bool:
    return (
        (isinstance(jsonable, Sequence) and not isinstance(jsonable, str))
        or isinstance(jsonable, Mapping)
    ) and len(jsonable) > 0
